# sage_setup: distribution = sagemath-modules
r"""
Type ``groups.affine.<tab>`` to access examples
of groups implemented as affine groups.
"""

# groups imported here will be available
# via  groups.affine.<tab>
#
# Do not use this file for code
#
# If you import a new group, then add an
# entry to the list in the module-level
# docstring of groups/groups_catalog.py

from .affine_group import AffineGroup as Affine
from .euclidean_group import EuclideanGroup as Euclidean
