# sage_setup: distribution = sagemath-flint
# Deprecated header file; use sage/libs/flint/mag.pxd instead
# See https://github.com/sagemath/sage/pull/36449

from sage.libs.flint.types cimport mag_t

from sage.libs.flint.mag cimport (
    MAG_BITS,
    mag_init,
    mag_clear,
    mag_init_set,
    mag_swap,
    mag_set,
    mag_zero,
    mag_one,
    mag_inf,
    mag_is_special,
    mag_is_zero,
    mag_is_inf,
    mag_is_finite,
    mag_equal,
    mag_cmp,
    mag_cmp_2exp_si,
    mag_min,
    mag_max,
    mag_print,
    mag_randtest,
    mag_randtest_special,
    mag_set_d,
    mag_set_ui,
    mag_set_fmpz,
    mag_set_d_2exp_fmpz,
    mag_set_fmpz_2exp_fmpz,
    mag_set_ui_2exp_si,
    mag_get_fmpq,
    mag_set_ui_lower,
    mag_set_fmpz_lower,
    mag_set_fmpz_2exp_fmpz_lower,
    mag_mul_2exp_si,
    mag_mul_2exp_fmpz,
    mag_mul,
    mag_mul_ui,
    mag_mul_fmpz,
    mag_add,
    mag_addmul,
    mag_add_2exp_fmpz,
    mag_div,
    mag_div_ui,
    mag_div_fmpz,
    mag_mul_lower,
    mag_mul_ui_lower,
    mag_mul_fmpz_lower,
    mag_add_lower,
    mag_sub_lower,
    mag_fast_init_set,
    mag_fast_zero,
    mag_fast_is_zero,
    mag_fast_mul,
    mag_fast_addmul,
    mag_fast_add_2exp_si,
    mag_fast_mul_2exp_si,
    mag_pow_ui,
    mag_pow_fmpz,
    mag_pow_ui_lower,
    mag_sqrt,
    mag_rsqrt,
    mag_hypot,
    mag_log1p,
    mag_log_ui,
    mag_exp,
    mag_expm1,
    mag_exp_tail,
    mag_binpow_uiui,
    mag_fac_ui,
    mag_rfac_ui,
    mag_bernoulli_div_fac_ui,
    mag_polylog_tail)
