# sage_setup: distribution = sagemath-objects
from typing import Any

class ParentWithBase:
    def __init__(self, base: Any, *args: Any, **kwds: Any) -> None: ...
    def base_extend(self, X: Any) -> Any: ...

def check_old_coerce(p: Any) -> None: ...
