# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/padic_mat.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    fmpz_mat_struct * padic_mat(const padic_mat_t A) noexcept
    fmpz * padic_mat_entry(const padic_mat_t A, slong i, slong j) noexcept
    slong padic_mat_val(const padic_mat_t A) noexcept
    slong padic_mat_prec(const padic_mat_t A) noexcept
    slong padic_mat_get_val(const padic_mat_t A) noexcept
    slong padic_mat_get_prec(const padic_mat_t A) noexcept
    slong padic_mat_nrows(const padic_mat_t A) noexcept
    slong padic_mat_ncols(const padic_mat_t A) noexcept
    void padic_mat_init(padic_mat_t A, slong r, slong c) noexcept
    void padic_mat_init2(padic_mat_t A, slong r, slong c, slong prec) noexcept
    void padic_mat_clear(padic_mat_t A) noexcept
    void _padic_mat_canonicalise(padic_mat_t A, const padic_ctx_t ctx) noexcept
    void _padic_mat_reduce(padic_mat_t A, const padic_ctx_t ctx) noexcept
    void padic_mat_reduce(padic_mat_t A, const padic_ctx_t ctx) noexcept
    bint padic_mat_is_empty(const padic_mat_t A) noexcept
    bint padic_mat_is_square(const padic_mat_t A) noexcept
    bint padic_mat_is_canonical(const padic_mat_t A, const padic_ctx_t p) noexcept
    void padic_mat_set(padic_mat_t B, const padic_mat_t A, const padic_ctx_t p) noexcept
    void padic_mat_swap(padic_mat_t A, padic_mat_t B) noexcept
    void padic_mat_swap_entrywise(padic_mat_t mat1, padic_mat_t mat2) noexcept
    void padic_mat_zero(padic_mat_t A) noexcept
    void padic_mat_one(padic_mat_t A) noexcept
    void padic_mat_set_fmpq_mat(padic_mat_t B, const fmpq_mat_t A, const padic_ctx_t ctx) noexcept
    void padic_mat_get_fmpq_mat(fmpq_mat_t B, const padic_mat_t A, const padic_ctx_t ctx) noexcept
    void padic_mat_get_entry_padic(padic_t rop, const padic_mat_t op, slong i, slong j, const padic_ctx_t ctx) noexcept
    void padic_mat_set_entry_padic(padic_mat_t rop, slong i, slong j, const padic_t op, const padic_ctx_t ctx) noexcept
    bint padic_mat_equal(const padic_mat_t A, const padic_mat_t B) noexcept
    bint padic_mat_is_zero(const padic_mat_t A) noexcept
    int padic_mat_fprint(FILE * file, const padic_mat_t A, const padic_ctx_t ctx) noexcept
    int padic_mat_fprint_pretty(FILE * file, const padic_mat_t A, const padic_ctx_t ctx) noexcept
    int padic_mat_print(const padic_mat_t A, const padic_ctx_t ctx) noexcept
    int padic_mat_print_pretty(const padic_mat_t A, const padic_ctx_t ctx) noexcept
    void padic_mat_randtest(padic_mat_t A, flint_rand_t state, const padic_ctx_t ctx) noexcept
    void padic_mat_transpose(padic_mat_t B, const padic_mat_t A) noexcept
    void _padic_mat_add(padic_mat_t C, const padic_mat_t A, const padic_mat_t B, const padic_ctx_t ctx) noexcept
    void padic_mat_add(padic_mat_t C, const padic_mat_t A, const padic_mat_t B, const padic_ctx_t ctx) noexcept
    void _padic_mat_sub(padic_mat_t C, const padic_mat_t A, const padic_mat_t B, const padic_ctx_t ctx) noexcept
    void padic_mat_sub(padic_mat_t C, const padic_mat_t A, const padic_mat_t B, const padic_ctx_t ctx) noexcept
    void _padic_mat_neg(padic_mat_t B, const padic_mat_t A) noexcept
    void padic_mat_neg(padic_mat_t B, const padic_mat_t A, const padic_ctx_t ctx) noexcept
    void _padic_mat_scalar_mul_padic(padic_mat_t B, const padic_mat_t A, const padic_t c, const padic_ctx_t ctx) noexcept
    void padic_mat_scalar_mul_padic(padic_mat_t B, const padic_mat_t A, const padic_t c, const padic_ctx_t ctx) noexcept
    void _padic_mat_scalar_mul_fmpz(padic_mat_t B, const padic_mat_t A, const fmpz_t c, const padic_ctx_t ctx) noexcept
    void padic_mat_scalar_mul_fmpz(padic_mat_t B, const padic_mat_t A, const fmpz_t c, const padic_ctx_t ctx) noexcept
    void padic_mat_scalar_div_fmpz(padic_mat_t B, const padic_mat_t A, const fmpz_t c, const padic_ctx_t ctx) noexcept
    void _padic_mat_mul(padic_mat_t C, const padic_mat_t A, const padic_mat_t B, const padic_ctx_t ctx) noexcept
    void padic_mat_mul(padic_mat_t C, const padic_mat_t A, const padic_mat_t B, const padic_ctx_t ctx) noexcept
