# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/partitions.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void partitions_rademacher_bound(arf_t b, const fmpz_t n, ulong N) noexcept
    void partitions_hrr_sum_arb(arb_t x, const fmpz_t n, slong N0, slong N, int use_doubles) noexcept
    void partitions_fmpz_fmpz(fmpz_t p, const fmpz_t n, int use_doubles) noexcept
    void partitions_fmpz_ui(fmpz_t p, ulong n) noexcept
    void partitions_fmpz_ui_using_doubles(fmpz_t p, ulong n) noexcept
    void partitions_leading_fmpz(arb_t res, const fmpz_t n, slong prec) noexcept
