# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/calcium.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    const char * calcium_version() noexcept
    ulong calcium_fmpz_hash(const fmpz_t x) noexcept
    void calcium_stream_init_file(calcium_stream_t out, FILE * fp) noexcept
    void calcium_stream_init_str(calcium_stream_t out) noexcept
    void calcium_write(calcium_stream_t out, const char * s) noexcept
    void calcium_write_free(calcium_stream_t out, char * s) noexcept
    void calcium_write_si(calcium_stream_t out, slong x) noexcept
    void calcium_write_fmpz(calcium_stream_t out, const fmpz_t x) noexcept
    void calcium_write_arb(calcium_stream_t out, const arb_t z, slong digits, ulong flags) noexcept
    void calcium_write_acb(calcium_stream_t out, const acb_t z, slong digits, ulong flags) noexcept
