# sage_setup: distribution = sagemath-modules
import builtins
from typing import Any

class CliffordAlgebraElement:
    def _repr_(self) -> str: ...
    def _latex_(self) -> str: ...
    def list(self) -> builtins.list[tuple[Any, Any]]: ...
    def support(self) -> builtins.list[Any]: ...
    def reflection(self) -> CliffordAlgebraElement: ...
    degree_negation = reflection
    def transpose(self) -> CliffordAlgebraElement: ...
    def conjugate(self) -> CliffordAlgebraElement: ...
    clifford_conjugate = conjugate

class ExteriorAlgebraElement(CliffordAlgebraElement):
    def reduce(self, I: Any, left: bool = True) -> ExteriorAlgebraElement: ...
    def interior_product(self, x: Any) -> ExteriorAlgebraElement: ...
    antiderivation = interior_product
    def hodge_dual(self) -> ExteriorAlgebraElement: ...
    def constant_coefficient(self) -> Any: ...
    def scalar(self, other: ExteriorAlgebraElement) -> Any: ...

class CohomologyRAAGElement(CliffordAlgebraElement): ...
