# sage_setup: distribution = sagemath-flint
# Deprecated header file; use sage/libs/flint/arf.pxd instead
# See https://github.com/sagemath/sage/pull/36449

from sage.libs.flint.types cimport arf_t

from sage.libs.flint.arf cimport (
    arf_init,
    arf_clear,
    arf_zero,
    arf_one,
    arf_pos_inf,
    arf_neg_inf,
    arf_nan,
    arf_is_zero,
    arf_is_one,
    arf_is_pos_inf,
    arf_is_neg_inf,
    arf_is_nan,
    arf_is_inf,
    arf_is_normal,
    arf_is_special,
    arf_is_finite,
    arf_set,
    arf_set_mpz,
    arf_set_fmpz,
    arf_set_ui,
    arf_set_si,
    arf_set_mpfr,
    arf_set_d,
    arf_swap,
    arf_init_set_ui,
    arf_init_set_si,
    arf_set_round,
    arf_set_round_si,
    arf_set_round_ui,
    arf_set_round_mpz,
    arf_set_round_fmpz,
    arf_set_si_2exp_si,
    arf_set_ui_2exp_si,
    arf_set_fmpz_2exp,
    arf_set_round_fmpz_2exp,
    arf_get_fmpz_2exp,
    arf_get_d,
    arf_get_mpfr,
    arf_get_fmpz,
    arf_get_si,
    arf_get_fmpz_fixed_fmpz,
    arf_get_fmpz_fixed_si,
    arf_floor,
    arf_ceil,
    arf_equal,
    arf_equal_si,
    arf_cmp,
    arf_cmpabs,
    arf_cmpabs_ui,
    arf_cmpabs_mag,
    arf_cmp_2exp_si,
    arf_cmpabs_2exp_si,
    arf_sgn,
    arf_min,
    arf_max,
    arf_bits,
    arf_is_int,
    arf_is_int_2exp_si,
    arf_abs_bound_lt_2exp_fmpz,
    arf_abs_bound_le_2exp_fmpz,
    arf_abs_bound_lt_2exp_si,
    arf_get_mag,
    arf_get_mag_lower,
    arf_set_mag,
    mag_init_set_arf,
    mag_fast_init_set_arf,
    arf_mag_set_ulp,
    arf_mag_add_ulp,
    arf_mag_fast_add_ulp,
    arf_init_set_shallow,
    arf_init_set_mag_shallow,
    arf_init_neg_shallow,
    arf_init_neg_mag_shallow,
    arf_randtest,
    arf_randtest_not_zero,
    arf_randtest_special,
    arf_debug,
    arf_print,
    arf_printd,
    arf_abs,
    arf_neg,
    arf_neg_round,
    arf_mul_2exp_si,
    arf_mul_2exp_fmpz,
    arf_mul,
    arf_mul_ui,
    arf_mul_si,
    arf_mul_mpz,
    arf_mul_fmpz,
    arf_add,
    arf_add_si,
    arf_add_ui,
    arf_add_fmpz,
    arf_add_fmpz_2exp,
    arf_sub,
    arf_sub_si,
    arf_sub_ui,
    arf_sub_fmpz,
    arf_addmul,
    arf_addmul_ui,
    arf_addmul_si,
    arf_addmul_mpz,
    arf_addmul_fmpz,
    arf_submul,
    arf_submul_ui,
    arf_submul_si,
    arf_submul_mpz,
    arf_submul_fmpz,
    arf_sum,
    arf_div,
    arf_div_ui,
    arf_ui_div,
    arf_div_si,
    arf_si_div,
    arf_div_fmpz,
    arf_fmpz_div,
    arf_fmpz_div_fmpz,
    arf_sqrt,
    arf_sqrt_ui,
    arf_sqrt_fmpz,
    arf_rsqrt,
    arf_complex_mul,
    arf_complex_mul_fallback,
    arf_complex_sqr)
