# sage_setup: distribution = sagemath-modules
from typing import Optional
from sage.matrix.matrix import Matrix

def reversed_hermite_form(mat: Matrix, transformation: bool = False) -> Optional[Matrix]:
    ...
