# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/perm.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    slong * _perm_init(slong n) noexcept
    void _perm_clear(slong * vec) noexcept
    void _perm_set(slong * res, const slong * vec, slong n) noexcept
    void _perm_one(slong * vec, slong n) noexcept
    void _perm_inv(slong * res, const slong * vec, slong n) noexcept
    void _perm_compose(slong * res, const slong * vec1, const slong * vec2, slong n) noexcept
    int _perm_parity(const slong * vec, slong n) noexcept
    int _perm_randtest(slong * vec, slong n, flint_rand_t state) noexcept
    int _perm_print(const slong * vec, slong n) noexcept
