# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/double_interval.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    di_t di_interval(double a, double b) noexcept
    di_t arb_get_di(const arb_t x) noexcept
    void arb_set_di(arb_t res, di_t x, slong prec) noexcept
    void di_print(di_t x) noexcept
    double d_randtest2(flint_rand_t state) noexcept
    di_t di_randtest(flint_rand_t state) noexcept
    di_t di_neg(di_t x) noexcept
    di_t di_fast_add(di_t x, di_t y) noexcept
    di_t di_fast_sub(di_t x, di_t y) noexcept
    di_t di_fast_mul(di_t x, di_t y) noexcept
    di_t di_fast_div(di_t x, di_t y) noexcept
    di_t di_fast_sqr(di_t x) noexcept
    di_t di_fast_add_d(di_t x, double y) noexcept
    di_t di_fast_sub_d(di_t x, double y) noexcept
    di_t di_fast_mul_d(di_t x, double y) noexcept
    di_t di_fast_div_d(di_t x, double y) noexcept
    di_t di_fast_log_nonnegative(di_t x) noexcept
    di_t di_fast_mid(di_t x) noexcept
    double di_fast_ubound_radius(di_t x) noexcept
