# sage_setup: distribution = sagemath-schemes
from sage.modular.hecke.homspace import HeckeModuleHomspace

from sage.modular.hecke.module import HeckeModule_free_module, HeckeModule_generic

from sage.modular.hecke.hecke_operator import HeckeOperator

from sage.modular.hecke.degenmap import DegeneracyMap

from sage.modular.hecke.algebra import HeckeAlgebra

from sage.modular.hecke.morphism import (HeckeModuleMorphism,
                                         HeckeModuleMorphism_matrix)

from sage.modular.hecke.element import HeckeModuleElement

from sage.modular.hecke.submodule import HeckeSubmodule

from sage.modular.hecke.ambient_module import AmbientHeckeModule
