# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/long_extras.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    size_t z_sizeinbase(slong n, int b) noexcept
    int z_mul_checked(slong * a, slong b, slong c) noexcept
    mp_limb_signed_t z_randtest(flint_rand_t state) noexcept
    mp_limb_signed_t z_randtest_not_zero(flint_rand_t state) noexcept
    mp_limb_signed_t z_randint(flint_rand_t state, mp_limb_t limit) noexcept
    int z_kronecker(slong a, slong n) noexcept
