# sage_setup: distribution = sagemath-objects
from sage.libs.gmp.types cimport *
from sage.libs.gmp.random cimport *
from sage.libs.gmp.mpz cimport *
from sage.libs.gmp.mpq cimport *
from sage.libs.gmp.pylong cimport *
